/************************************************************************
* "Epsilon" is a registered trademark licensed to Lugaru Software, Ltd. *
*		"EEL" and "Lugaru" are trademarks of Lugaru Software, Ltd.		*
*																		*
*  Copyright (C) 1997, 2003 Lugaru Software Ltd.  All rights reserved.	*
*																		*
* Limited permission is hereby granted to reproduce and modify this		*
* copyrighted material provided that the resulting code is used only in *
* conjunction with Lugaru products and that this notice is retained in	*
* any such reproduction or modification.								*
************************************************************************/

// Updated for v12.

#include "eel.h"
#include "c.h"
#include "tags.h"
#include "idcomp.h"

is_identifier_char(int c)	// Could this char be part of an identifier?
{
	return c >= 0 && IS_ASCII_KEY(c) && (isident(c) || c >= 0x80);
}

// Called after inserting a character.  Perform auto-completion, then
// discard/keep the completion after getting another key.

new_ident_complete_normal_character()
{
	int old_class, keep = 0, cmd_index;

	ident_complete_normal_character();
	if (!auto_complete || char_avail() || !is_identifier_char(key))
		return;
	save_spot mark;
	if (!auto_complete_identifier())
		return;			// No completion to do.
	old_class = modify_region(1, MRCOLOR, color_class ident_completion);
	maybe_refresh();	// Show the completion we did.
	save_var permit_window_keys = PERMIT_SCROLL_KEY | PERMIT_WHEEL_KEY;
	getkey();			// Wait for next user action.
	while (key == MOUSE_MOVE) {
		do_again();		// Mouse moves don't count, just process them.
		getkey();
		maybe_refresh();
	}
	modify_region(1, MRCOLOR, old_class);
	cmd_index = (key & CMD_INDEX_KEY) ? -1 : reg_tab[key];
	if (cmd_index == complete_identifier)
		keep = 1;		// Keep the completion, then do the key.
	else if (IS_ENTER_KEY(key) || key == '\t')
		keep = 2;		// Keep and eat this key.
	else if (key == CTRL('H') || cmd_index == delete_character)
		keep = -1;		// Discard the completion.
	else if (auto_complete == 1 && (!IS_ASCII_KEY(key) || IS_CTRL_KEY(key)))
		keep = 1;		// Keep the completion, then do the key.
	if (keep > 0) {				// Go past and remove highlight.
		undo_to_flag(UNDO_FLAG_POST_COMPL);
		highlight_off();
		if (keep != 2)
			ungot_key = full_key;
	} else {				// Discard inserted text and process key.
		undo_to_flag(UNDO_FLAG_IDENT_COMPL);
		highlight_off();
		this_cmd = CMD_ID_COMPL;
		if (!keep)
			ungot_key = full_key;
	}
}

REPLACE_FUNC("ident-complete", "normal-character")

// Auto-complete this identifier if possible, return nonzero if we did.

int auto_complete_identifier()
{
	char word[FNAMELEN], biggest[FNAMELEN], id[FNAMELEN];
	int count;

	if (!completion_sources || !is_identifier_char(character(point - 1))
		|| is_identifier_char(curchar()) || is_highlight_on()
		|| window_to_screen(window_handle)
		|| !get_mode_variable("%s-ok-to-complete"))
		return 0;	// Don't do anything.
	word[0] = '\n';		// Get the partial word with a \n before it.
	if (!parse_string(-1, IDENT_PATTERN, word + 1))
		return 0;
	point = matchstart;
	if (strlen(word + 1) < completion_minimum_start || isdigit(word[1]))
		return 0;
	*biggest = 0;
	if (!look_for_some_completions(word, biggest, 0) || !*biggest)
		return 0;
	if (prev_cmd != CMD_ID_COMPL && prev_cmd != CMD_SHOW_COMPL
		&& strlen(biggest) < completion_minimum_add)
		return 0;			// Not enough would be added.
	if (auto_complete == 3) {		// Just display a message.
		count = make_completion_list(word);
		find_buffer_prefix(name_to_bufnum(_MATCH_BUF), id);
		if (strlen(id) <= strlen(word) - 1 && prev_cmd != CMD_SHOW_COMPL)
			return 0;
		if (count == 1)
			note("Alt-= changes %s => %s", word + 1, id);
		else
			note("Alt-= changes %s => %s [%d matches]", word + 1, id, count);
		expire_message = 1;
		this_cmd = CMD_SHOW_COMPL;
		return 0;
	}
	undo_flag = UNDO_FLAG_IDENT_COMPL;
	undo_mainloop();
	mark = point;
	set_region_type(REGNORM);
	stuff(biggest);
	undo_flag = UNDO_FLAG_POST_COMPL;
	highlight_on();
	return 1;
}

// Throw away all cached completions for this buffer.
discard_buffer_completions()
{
	char bname[FNAMELEN];

	sprintf(bname, "-completion-%s", bufname);
	delete_buffer(bname);
}

// Look in each source for a completion of word.  If list is nonzero,
// copy each match to the buffer named by dest.  If list is zero,
// copy the longest possible completion to the character array dest.
// Return 0 on error, or if there was no unique completion, 1 otherwise.

look_for_some_completions(char *word, char *dest, int list)
{
	char bname[FNAMELEN], func[FNAMELEN];
	int buf, i, orig = bufnum;

	save_var window_bufnum, bufnum, point, mark;
	if (completion_sources & IDENT_COMP_LOCALS) {
		sprintf(bname, "-completion-%s", bufname);
		if (!exist(bname)) {
			buf = create(bname);
			sprintf(func, "make_%s_completion_list", major_mode);
			i = find_index(func);	// Call mode-specific function.
			switch (name_type(i)) {		// i has name table index of subr
				case NT_SUBR: case NT_COMMAND:
				case NT_AUTOSUBR: case NT_AUTOLOAD:
					(*(int (*)())i)(buf);
					break;	// Cast i to a func ptr, then call it.
			}
		}
		if (!look_for_completion(bname, word, dest, "\n", list))
			return 0;
	}
	if (completion_sources & IDENT_COMP_TAGS) {
		if (!exist("-tags") && !exist("-notags")) {
			if (tags_available()) {
				init_tags();
				say("");
			} else
				zap("-notags");
		}
		if (exist("-tags"))
			if (!look_for_completion("-tags", word, dest, "\t", list))
				return 0;
	}

	if ((completion_sources & IDENT_COMP_STD))
		if (!look_for_file_completions(word, dest, list, IDENT_COMPL_STD_KWDS_BUF,
									   IDENT_COMPL_STD_KWDS, orig))
			return 0;
	if ((completion_sources & IDENT_COMP_FILE))
		if (!look_for_file_completions(word, dest, list, IDENT_COMPL_KWDS_BUF,
									   IDENT_COMPL_KWDS, orig))
			return 0;
	return 1;
}

// Look in the file for a completion of word from the buffer origbuf.
// Cache the file in the buffer buf.  If list is nonzero, copy each
// match to the buffer named by dest.  If list is zero, copy the
// longest possible completion to the character array dest.  Return -1
// if we couldn't find the file, 0 if there was no unique completion or
// on other error, 1 otherwise.

look_for_file_completions(char *word, char *dest, int list,
						  char *buf, char *file, int origbuf)
{
	char *s;

	if (!exist(buf)) {	// 1st time: read the file.
		bufnum = create(buf);
		if (!(s = lookpath(file)) || err_file_read(s, default_translation_type))
			return -1;
		filename = s;
		buffer_not_saveable = 1;
	}
	if (!look_for_conditional_completions(buf, word, dest, "\n", list, origbuf))
		return 0;
	return 1;
}

// Look in buffer bname for a completion of word (as found in the
// buffer origbuf), but interpret condition lines too.  Each block of
// possible completions may have a series of "=cond" condition lines
// before it; the block will be ignored unless the conditions are met.

// Each possible completion ends with the text in ender (normally \t or \n).
// Word must have a \n before it.
// If !list, return added chars in dest, modifying it, and return
// 0 if a completion would be ambiguous (or on error), 1 otherwise.
// If list, copy matching words to the buffer named dest.

int look_for_conditional_completions(char *bname, char *word, char *dest,
									 char *ender, int list, int origbuf)
{
	char msg[FNAMELEN];
	int do_block;

	save_var bufname = bname;
	save_var narrow_end = 0, narrow_start = 0, point = 0;
	save_var case_fold = 0;
	for (;;) {
		do_block = 1;		// Assume we want to look in the next block.
		while (curchar() == '=') {
			if (parse_string(1, "=cond[ \t]+")) {
				point = matchend;
				switch (interpret_completion_conditional(origbuf)) {
					case 0:				// Skip this block.
						do_block = 0;	// (fall through)
					case 1:				// Yes, do the block.
						if (curchar() == '\n')
							break;	// Parsed it all, it's good.
					default:
						sprintf(msg, "%s line %d, position %d: "
								"Error in completion conditional",
								filename, lines_between(0, point, 0) + 1,
								point);
						say("%s", msg);
						buffer_printf("completion-conditional-errors", "%s\n",
									  msg);
						break;
				}
			}
			if (!nl_forward())
				break;
		}
		narrow_start = point;
		if (search(1, "\n=cond"))
			to_begin_line();
		if (point == narrow_start)	// No more.
			break;
		narrow_end = size() - point;
		if (do_block)
			if (!look_for_completion(bname, word, dest, ender, list))
				return 0;
		point = size();
		narrow_end = narrow_start = 0;
	}
	return 1;
}

// Is the conditional completion expression that follows true?
// Return 1 if true, 0 if false, or return -1 if syntax error.

int interpret_completion_conditional(int origbuf)
{
	char varname[FNAMELEN], val[FNAMELEN], true, c, *p;
	int i, type, a, b, invert;

	if (!parse_string(1, "[-_a-z0-9]+", varname))
		return -1;
	point += parse_string(1, "[ \t]*");
	if (invert = (parse_string(1, "%!&") != 0))
		point++;
	if (!parse_string(1, "([!=]=|&)[ \t]*"))
		return -1;
	true = (c = character(matchstart)) == '=';
	point = matchend;
	type = vartype(i = find_index(varname));
	switch (type) {
		case TYPE_CHAR:
		case TYPE_SHORT:
		case TYPE_INT:
			if (!parse_string(1, "[0-9][0-9oxb]*", val))
				return -1;
			save_var bufnum = origbuf;
			a = get_num_var(i);
			restore_vars();
			b = numtoi(val);
			if (c == '&')		// bitwise and
				true = (a & b) ? !invert : invert;
			else if (a != b)	// == or !=
				true = !true;
			break;
		case TYPE_CPTR:
		case TYPE_CARRAY:
			if (!parse_string(1, "\"[^\"\n]*\"", val))
				return -1;
			val[strlen(val) - 1] = 0;	// Remove final " char.
			if (!strchr("!=", c))	// Only == or != allowed.
				return -1;
			save_var bufnum = origbuf;
			p = get_str_var(i);
			restore_vars();
			if (strcmp(p, val + 1))
				true = !true;
			break;
		default:
			return -1;
	}
	point += parse_string(1, "[ \t]*");
	if (parse_string(1, "%|%|[ \t]*", val))
		true = interpret_completion_conditional(origbuf) || true;
	if (parse_string(1, "&&[ \t]*", val))
		true = interpret_completion_conditional(origbuf) && true;
	return true;
}


// Modify dest by chopping off any suffix that doesn't match word.
// Return 0 if dest ever becomes empty (or word is empty), else 1.
// Special case: If dest is empty on entry, then this is the first
// time, so initialize dest with word.

shrink_completion(char *dest, char *word)
{
	int i;

	if (!*word)		// Word's in the list as typed, don't add more to it.
		return 0;
	if (!*dest)		// First match we found, begin with this.
		strcpy(dest, word);
	else {			// Trim the existing completion.
		for (i = 0; word[0]; i++)
			if (charfcmp(word[i], dest[i]))
				break;
		dest[i] = 0;
		if (!i)		// Nothing left in the completion.
			return 0;
	}
	return 1;
}

// Look in buffer bname for a completion of word.  Each possible
// completion ends with the text in ender (normally \t or \n).
// Word must have a \n before it.
// If !list, return added chars in dest, modifying it, and return
// 0 if a completion would be ambiguous (or on error), 1 otherwise.
// If list, copy matching words to the buffer named dest.

int look_for_completion(char *bname, char *word, char *dest, char *ender,
					   int list)
{
	char this[FNAMELEN];
	int i;

	save_var bufname = bname;
	point = 0;
	save_var case_fold = 0;
	while (search(1, word)) {
		if (list) {		// Get the entire match.
			i = give_begin_line();
			search(1, ender);
			xfer(dest, i - 1, point - 1);
			point--;
			continue;
		}
		i = point;
		search(1, ender);	// Find the end of the completion.
		if (point - i > FNAMELEN - 3) {
			say("Bad format: identifier too long near line %d in %s",
				lines_between(0, i, 1) + 1, bufname);
			return 0;
		}
		point--;		// Get the rest of the completion.
		grab(i, point, this);
		if (!shrink_completion(dest, this))
			return 0;
	}
	return 1;
}

// Find the longest prefix shared by all lines in this buffer after point.
find_buffer_prefix(int buf, char *prefix)
{
	char next[FNAMELEN];
	int i;

	save_var bufnum = buf;
	grab_line(buf, prefix);
	while (grab_line(buf, next)) {
		for (i = 0; prefix[i]; i++)
			if (prefix[i] != next[i])
				break;
		prefix[i] = 0;
	}
}

// Make list of matching words in buffer _MATCH_BUF, return count.
make_completion_list(char *word)
{
	zap(_MATCH_BUF);
	look_for_some_completions(word, _MATCH_BUF, 1);
	save_var bufname = _MATCH_BUF;
	save_var sort_status = 0;
	do_sort_region(0, size(), 0, 0);
	point = 0;		// List of completions always starts with a blank line.
	do_uniq(1, 1, 0);
	delete_final_newlines();
	point = 1;		// Start seaching list from non-blank line.
	return lines_between(0, size(), 0);
}

command complete_identifier() on reg_tab[ALT('=')]
{
	char word[FNAMELEN], id[FNAMELEN], *s;
	int orig = bufnum, wstart, wend, count;

	if (_doing_ident_completion) {	// Just list the matches again.
		ungot_key = '?';
		return;
	}
	save_var bufnum, point;
	iter = 0;
	if (has_arg)		// Delete all completion buffers, rebuild later.
		while (s = buf_match("-completion-", STARTMATCH))
			delete_buffer(s);
	discard_buffer_completions();
	if (is_window(window_handle) != ISTILED)
		error("Can't complete here.");
	bufnum = orig;
	if (is_identifier_char(curchar()))		// To end of word.
		point += parse_string(1, IDENT_PATTERN);
	strcpy(word, "\n");		// Word to complete will always start with \n.
	wend = point;
	wstart = point -= parse_string(-1, IDENT_PATTERN, word + 1);
	if (strlen(word + 1) < 1)
		error("Can't complete here.");
	count = make_completion_list(word);
	restore_vars();
	find_buffer_prefix(name_to_bufnum(_MATCH_BUF), id);
	if (strlen(id) <= wend - wstart) {	// Must ask.
		ungot_key = '?';
		save_var insert_default_response = 1, _doing_ident_completion = 1;
		comp_read(id, "Select an identifier to insert: ",
				  general_matcher, MUST_MATCH | NONE_OK, word + 1);
	} else {
		if (count == 1)	 // Found something to insert w/o asking.
			say("%s => %s", word + 1, id);
		else
			say("%s => %s [%d matches]", word + 1, id, count);
		expire_message = 1;
	}
	point = wend;
	delete(wstart, wend);
	stuff(id);
}

/* Make in buffer ids a list of all identifiers in the current C-mode
buffer. */

make_c_completion_list(int ids)
{
	int p, c, talk = size() > 200000;
	char buf[4], exclude = point;

	if (talk)
		sayput("Making completion list, one moment...");
	buf_zap(ids);
	buf_xfer(ids, 0, size());
	save_var bufnum = ids;
	buffer_not_saveable = discardable_buffer = 1;
	save_var case_fold = 0;	// Don't confuse STRUCT with struct.
	save_var abort_searching = 0;
	point = exclude;		// Don't include word at point.
	delete(point - parse_string(-1, word_pattern),
		   point + parse_string(1, word_pattern));
	point = 0;
	strcpy(buf, "X|\\");
	while (re_search(1, "/<*|/>|[\"']")) {
		p = point;		// strip all comments and strings from buffer
		switch (c = character(point - 1)) {
			case '"':
			case '\'':	buf[0] = c;		// string or char constant
						while (re_search(1, buf)
							   && character(point - 1) == '\\')
							point++;
						delete(p - 1, point);
						break;
			case '/':	delete(p - 2, give_end_line());	// one-line comment
						break;
			case '*':	search(1, "*/");		// block comment
						delete(p - 2, point);
						break;
		}
		insert(' ');		// replace it with a single space
	}
	simple_re_replace(1, "[^a-zA-Z0-9_]+", "\n"); // delete all but identifiers
	save_var sort_status = 0;
	do_sort_region(0, size(), 0, 0);
	point = 0;
	do_uniq(1, 1, 0);
	point = 0;
	re_search(1, "^![^0-9\n]");
	delete(0, point - 1);
	if (talk)
		say("");
}

// Mode-specific routine: Should auto-completion happen here?

c_ok_to_complete()
{
	return !(in_c_comment(c_look_back) & (IN_STRING | IN_COMMENT));
}
